import asyncio
import logging
from asyncio import StreamReader, StreamWriter, IncompleteReadError, Task, CancelledError
from typing import Optional, Coroutine, Callable

from aioaprs.config import AioAPRSClientConfig
from aioaprs.version import __VERSION__

_logger = logging.getLogger(__name__)


class AioAPRSClient(object):
    _lock: asyncio.Lock
    _keep_running: bool
    _task_rx: Optional[Task]

    _config: AioAPRSClientConfig
    _callback: Optional[Callable[[str], Coroutine]]

    _sock_reader: Optional[StreamReader]
    _sock_writer: Optional[StreamWriter]

    def __init__(self, config: AioAPRSClientConfig, callback: Optional[Callable[[str], Coroutine]] = None) -> None:
        super().__init__()

        self._lock = asyncio.Lock()
        self._keep_running = False
        self._task_rx = None

        self._config = config
        self._callback = callback

        self._sock_reader = None
        self._sock_writer = None

    async def connect(self) -> None:
        async with self._lock:
            if self._sock_reader is not None \
                    or self._sock_writer is not None \
                    or self._task_rx is not None:
                return

            self._sock_reader, self._sock_writer = await asyncio.open_connection(
                host=self._config.host,
                port=self._config.port
            )

            self._keep_running = True
            self._task_rx = asyncio.create_task(self._loop())

            password: str = self.passcode(self._config.callsign)
            version: str = f"aioaprs {__VERSION__}"

            filter_string: str = ""
            if self._config.server_filter:
                filter_string = f" filter {self._config.server_filter}"

            login_line: str = f"user {self._config.callsign} pass {password} vers {version}{filter_string}"
            await self._write(login_line)

    async def close(self) -> None:
        async with self._lock:
            self._keep_running = False

            if self._task_rx is not None:
                self._task_rx.cancel()
                self._task_rx = None

            if self._sock_reader is not None:
                self._sock_reader = None

            if self._sock_writer is not None:
                self._sock_writer.close()
                self._sock_writer = None

    async def gather(self) -> None:
        try:
            await asyncio.gather(self._task_rx)
        except CancelledError:
            pass

    async def _write(self, data: str) -> None:
        _logger.debug(f"[     TX >>> ] {data}")

        rawdata: bytes = data.encode() + b"\r\n"
        self._sock_writer.write(rawdata)
        await self._sock_writer.drain()

    async def _loop(self) -> None:
        while self._keep_running:
            try:
                raw_line: bytes = await self._sock_reader.readuntil()
            except CancelledError:
                break
            except IncompleteReadError:
                break

            line: str = raw_line.decode("utf-8", "ignore").strip()
            if not line:
                break

            _logger.debug(f"[ <<< RX     ] {line}")

            if not line.startswith("#") and self._callback is not None:
                asyncio.create_task(self._callback(line))

    @staticmethod
    def passcode(callsign: str) -> str:
        callsign = callsign.split("-")[0].upper()

        code = 0x73e2
        for i, char in enumerate(callsign):
            code ^= ord(char) << (8 if not i % 2 else 0)

        return "%d" % (code & 0x7fff)

    @staticmethod
    def version() -> str:
        return __VERSION__



from aioaprs.packets import PacketType
from aioaprs.serializers.telemetry import serialize_body_telemetry


class PacketSerializer:
    _packet: dict

    def __init__(self, packet: dict) -> None:
        super().__init__()

        self._packet = packet

    def serialize(self) -> str:
        if self._packet is None:
            raise ValueError("Invalid packet")

        if not self._packet:
            raise ValueError("Empty packet")

        return "%s:%s" % (
            self._serialize_header(),
            self._serialize_body(),
        )

    def _serialize_header(self) -> str:
        header = "%s>%s" % (self._packet["source"], self._packet["destination"])

        if "path" in self._packet and isinstance(self._packet["path"], list) and self._packet["path"]:
            header += "," + ",".join(self._packet["path"])
        elif "via" in self._packet and isinstance(self._packet["via"], str) and self._packet["via"]:
            header += "," + self._packet["via"]

        return header

    def _serialize_body(self) -> str:
        if self._packet["type"] == PacketType.TELEMETRY_DATA:
            return serialize_body_telemetry(self._packet)

        else:
            return "body" in self._packet and self._packet["body"] or ""



import logging
from typing import Dict, List

from aioaprs.packets import PacketType

_logger = logging.getLogger(__name__)


class TelemetryEnricher:
    _parameter_names: Dict[str, List[str]]
    _unit_labels: Dict[str, List[str]]
    _equation_coefficients: Dict[str, List[Dict[str, float]]]
    _project_names: Dict[str, str]

    def __init__(self) -> None:
        super().__init__()

        self._parameter_names = dict()
        self._unit_labels = dict()
        self._equation_coefficients = dict()
        self._project_names = dict()

    def parse(self, packet: dict) -> None:
        self._validate_packet(packet)

        if packet["type"] != PacketType.MESSAGE:
            raise ValueError("Packet type not supported for parsing")

        if packet["telemetry_data"]["type"] == "PARM":
            self._parameter_names[packet["addressee"]] = packet["telemetry_data"]["items"]

        elif packet["telemetry_data"]["type"] == "UNIT":
            self._unit_labels[packet["addressee"]] = packet["telemetry_data"]["items"]

        elif packet["telemetry_data"]["type"] == "EQNS":
            self._equation_coefficients[packet["addressee"]] = packet["telemetry_data"]["equations"]

        elif packet["telemetry_data"]["type"] == "BITS":
            self._project_names[packet["addressee"]] = packet["telemetry_data"]["project_name"]

    def enrich(self, packet: dict) -> None:
        self._validate_packet(packet)

        if packet["type"] != PacketType.TELEMETRY_DATA:
            raise ValueError("Packet type not supported for enriching")

        if packet["source"] in self._parameter_names:
            packet["parameter_names"] = self._parameter_names[packet["source"]]

        if packet["source"] in self._unit_labels:
            packet["unit_labels"] = self._unit_labels[packet["source"]]

        if packet["source"] in self._equation_coefficients:
            packet["equation_coefficients"] = self._equation_coefficients[packet["source"]]

            packet["values_real"] = []
            for pos in range(min(len(packet["values"]), 5)):
                value_raw: float = packet["values"][pos]
                a: float = packet["equation_coefficients"][pos]["a"]
                b: float = packet["equation_coefficients"][pos]["b"]
                c: float = packet["equation_coefficients"][pos]["c"]
                real_value: float = (a * value_raw * value_raw) + (b * value_raw) + c
                packet["values_real"].append(real_value)

        if packet["source"] in self._project_names:
            packet["project_name"] = self._project_names[packet["source"]]

    @staticmethod
    def _validate_packet(packet: dict) -> None:
        if packet is None:
            raise ValueError("Invalid packet")

        if not packet:
            raise ValueError("Empty packet")

        if "type" not in packet or packet["type"] not in [PacketType.TELEMETRY_DATA, PacketType.MESSAGE]:
            raise ValueError("Packet type not valid")



import enum


class PacketType(enum.Enum):
    POSITION_WITHOUT_TIMESTAMP = "Position without timestamp (no APRS messaging), or Ultimeter 2000 WX Station"
    POSITION_WITH_TIMESTAMP = "Position with timestamp (no APRS messaging)"
    POSITION_WITHOUT_TIMESTAMP_APRS = "Position without timestamp (with APRS messaging)"
    POSITION_WITH_TIMESTAMP_APRS = "Position with timestamp (with APRS messaging)"
    PEET_BROS_U_II_WEATHER_STATION = "Peet Bros U-II Weather Station"
    RAW_GPS_DATA = "Raw GPS data or Ultimeter 2000"
    MICROFINDER = "Agrelo DFJr / MicroFinder"
    MIC_E_DATA_OLD = "Old Mic-E Data (but Current data for TM-D700)"
    MIC_E_DATA = "Current Mic-E Data (not used in TM-D700)"
    ITEM = "Item"
    MESSAGE = "Message"
    OBJECT = "Object"
    STATION_CAPABILITIES = "Station Capabilities"
    STATUS = "Status"
    QUERY = "Query"
    TELEMETRY_DATA = "Telemetry data"
    MAIDENHEAD_BEACON = "Maidenhead grid locator beacon (obsolete)"
    WEATHER_REPORT = "Weather Report (without position)"
    USER_DEFINED = "User-Defined APRS packet format"
    THIRD_PARTY = "Third-party traffic"

    UNUSED = "[Unused]"
    DO_NOT_USE = "[Do not use]"
    DO_NOT_USE_TNC = "[Do not use - TNC stream switch character]"
    RESERVER_MAP_FEATURE = "[Reserved - Map Feature]"
    RESERVER_SHELTER_DATA_WITH_TIME = "[Reserved - Shelter data with time]"
    RESERVER_SPACE_WEATHER = "[Reserved - Space weather]"
    INVALID_OR_TEST = "Invalid data or test data"


DATA_TYPE_IDENTIFIER: dict = {
    "!": PacketType.POSITION_WITHOUT_TIMESTAMP,
    "\"": PacketType.UNUSED,
    "#": PacketType.PEET_BROS_U_II_WEATHER_STATION,
    "$": PacketType.RAW_GPS_DATA,
    "%": PacketType.MICROFINDER,
    "&": PacketType.RESERVER_MAP_FEATURE,
    "'": PacketType.MIC_E_DATA_OLD,
    "(": PacketType.UNUSED,
    ")": PacketType.ITEM,
    "*": PacketType.PEET_BROS_U_II_WEATHER_STATION,
    "+": PacketType.RESERVER_SHELTER_DATA_WITH_TIME,
    ",": PacketType.INVALID_OR_TEST,
    "-": PacketType.UNUSED,
    ".": PacketType.RESERVER_SPACE_WEATHER,
    "/": PacketType.POSITION_WITH_TIMESTAMP,
    "0": PacketType.DO_NOT_USE,
    "1": PacketType.DO_NOT_USE,
    "2": PacketType.DO_NOT_USE,
    "3": PacketType.DO_NOT_USE,
    "4": PacketType.DO_NOT_USE,
    "5": PacketType.DO_NOT_USE,
    "6": PacketType.DO_NOT_USE,
    "7": PacketType.DO_NOT_USE,
    "8": PacketType.DO_NOT_USE,
    "9": PacketType.DO_NOT_USE,
    ":": PacketType.MESSAGE,
    ";": PacketType.OBJECT,
    "<": PacketType.STATION_CAPABILITIES,
    "=": PacketType.POSITION_WITHOUT_TIMESTAMP_APRS,
    ">": PacketType.STATUS,
    "?": PacketType.QUERY,
    "@": PacketType.POSITION_WITH_TIMESTAMP_APRS,
    "A": PacketType.DO_NOT_USE,
    "B": PacketType.DO_NOT_USE,
    "C": PacketType.DO_NOT_USE,
    "D": PacketType.DO_NOT_USE,
    "E": PacketType.DO_NOT_USE,
    "F": PacketType.DO_NOT_USE,
    "G": PacketType.DO_NOT_USE,
    "H": PacketType.DO_NOT_USE,
    "I": PacketType.DO_NOT_USE,
    "J": PacketType.DO_NOT_USE,
    "K": PacketType.DO_NOT_USE,
    "L": PacketType.DO_NOT_USE,
    "M": PacketType.DO_NOT_USE,
    "N": PacketType.DO_NOT_USE,
    "O": PacketType.DO_NOT_USE,
    "P": PacketType.DO_NOT_USE,
    "Q": PacketType.DO_NOT_USE,
    "R": PacketType.DO_NOT_USE,
    "S": PacketType.DO_NOT_USE,
    "T": PacketType.TELEMETRY_DATA,
    "U": PacketType.DO_NOT_USE,
    "V": PacketType.DO_NOT_USE,
    "W": PacketType.DO_NOT_USE,
    "X": PacketType.DO_NOT_USE,
    "Y": PacketType.DO_NOT_USE,
    "Z": PacketType.DO_NOT_USE,
    "[": PacketType.MAIDENHEAD_BEACON,
    "\\": PacketType.UNUSED,
    "]": PacketType.UNUSED,
    "^": PacketType.UNUSED,
    "_": PacketType.WEATHER_REPORT,
    "`": PacketType.MIC_E_DATA,
    "a": PacketType.DO_NOT_USE,
    "b": PacketType.DO_NOT_USE,
    "c": PacketType.DO_NOT_USE,
    "d": PacketType.DO_NOT_USE,
    "e": PacketType.DO_NOT_USE,
    "f": PacketType.DO_NOT_USE,
    "g": PacketType.DO_NOT_USE,
    "h": PacketType.DO_NOT_USE,
    "i": PacketType.DO_NOT_USE,
    "j": PacketType.DO_NOT_USE,
    "k": PacketType.DO_NOT_USE,
    "l": PacketType.DO_NOT_USE,
    "m": PacketType.DO_NOT_USE,
    "n": PacketType.DO_NOT_USE,
    "o": PacketType.DO_NOT_USE,
    "p": PacketType.DO_NOT_USE,
    "q": PacketType.DO_NOT_USE,
    "r": PacketType.DO_NOT_USE,
    "s": PacketType.DO_NOT_USE,
    "t": PacketType.DO_NOT_USE,
    "u": PacketType.DO_NOT_USE,
    "v": PacketType.DO_NOT_USE,
    "w": PacketType.DO_NOT_USE,
    "x": PacketType.DO_NOT_USE,
    "y": PacketType.DO_NOT_USE,
    "z": PacketType.DO_NOT_USE,
    "{": PacketType.USER_DEFINED,
    "|": PacketType.DO_NOT_USE_TNC,
    "}": PacketType.THIRD_PARTY,
    "~": PacketType.DO_NOT_USE_TNC
}

PACKET_TYPE_IDENTIFIER: dict = {
    PacketType.POSITION_WITHOUT_TIMESTAMP: "!",
    PacketType.RAW_GPS_DATA: "$",
    PacketType.MICROFINDER: "%",
    PacketType.RESERVER_MAP_FEATURE: "&",
    PacketType.MIC_E_DATA_OLD: "'",
    PacketType.ITEM: ")",
    PacketType.PEET_BROS_U_II_WEATHER_STATION: "*",
    PacketType.RESERVER_SHELTER_DATA_WITH_TIME: "+",
    PacketType.INVALID_OR_TEST: ",",
    PacketType.RESERVER_SPACE_WEATHER: ".",
    PacketType.POSITION_WITH_TIMESTAMP: "/",
    PacketType.MESSAGE: ":",
    PacketType.OBJECT: ";",
    PacketType.STATION_CAPABILITIES: "<",
    PacketType.POSITION_WITHOUT_TIMESTAMP_APRS: "=",
    PacketType.STATUS: ">",
    PacketType.QUERY: "?",
    PacketType.POSITION_WITH_TIMESTAMP_APRS: "@",
    PacketType.TELEMETRY_DATA: "T",
    PacketType.MAIDENHEAD_BEACON: "[",
    PacketType.WEATHER_REPORT: "_",
    PacketType.MIC_E_DATA: "`",
    PacketType.USER_DEFINED: "{",
    PacketType.DO_NOT_USE_TNC: "|",
    PacketType.THIRD_PARTY: "}",
    "~": PacketType.DO_NOT_USE_TNC
}



import logging
from typing import List

from aioaprs.packets import PacketType, PACKET_TYPE_IDENTIFIER

_logger = logging.getLogger(__name__)


def serialize_body_telemetry(packet: dict) -> str:
    serialized_packet: str = f"{PACKET_TYPE_IDENTIFIER[PacketType.TELEMETRY_DATA]}#"

    if 0 <= packet["seq"] <= 255:
        serialized_packet += "%03d" % packet["seq"]
    else:
        serialized_packet += "MIC"

    if "values" in packet and packet["values"]:
        serialized_packet += ","

        values: List[str] = []
        for value in packet["values"][:5]:
            values.append("%.02f" % value)

        serialized_packet += ",".join(values)

        if len(packet["values"]) >= 5 and "bits" in packet and packet["bits"]:
            serialized_packet += ","

            for bit in packet["bits"][:8]:
                serialized_packet += (bit is True) and "1" or "0"

    return serialized_packet



DEFAULT_PORT_HIGHSPEED: int = 10152
DEFAULT_PORT_FILTER: int = 14580

DEFAULT_HOST: str = "rotate.aprs.net"
DEFAULT_PORT: int = DEFAULT_PORT_HIGHSPEED
DEFAULT_CALLSIGN: str = "nocall"
DEFAULT_SERVER_FILTER: str = ""
DEFAULT_HEARTBEAT: int = 15



import logging
from typing import List

_logger = logging.getLogger(__name__)


def parse_body_telemetry(raw_body: str, strict_mode: bool = False) -> dict:
    packet: dict = dict()

    body_data = raw_body[1:]
    if body_data.startswith("#"):
        body_data = body_data[1:]

    items: List[str] = body_data.split(",")

    if len(items[0]) != 3:
        _logger.warning(f"Wrong value for seq number: {items[0]}")
        if strict_mode:
            raise ValueError(f"Invalid seq number: {items[0]}")

    packet["seq"] = int(items[0])
    packet["values"] = []

    raw_values_analog: List[str] = items[1:6]

    for value in raw_values_analog:
        float_value: float = 0.0

        try:
            float_value = float(value)
        except ValueError:
            _logger.error(f"Unable to convert \"{value}\" to float")

        packet["values"].append(float_value)

    if all("." not in x for x in raw_values_analog) and all(0 <= x <= 255 for x in packet["values"]):
        map(int, packet["values"])

    if len(items) >= 7 and set(items[6][:8]) <= set("01"):
        packet["bits"] = []

        for bit in items[6][:8]:
            bool_value: bool = False

            try:
                bool_value = bool(bit == "1")
            except ValueError:
                _logger.error(f"Unable to convert \"{bit}\" to float")

            packet["bits"].append(bool_value)

        while len(packet["bits"]) < 8:
            packet["bits"].append(False)

    return packet



from aioaprs.default import *


class AioAPRSClientConfig:
    _host: str
    _port: int
    _callsign: str
    _server_filter: str
    _heartbeat: int

    def __init__(self) -> None:
        super().__init__()

        self._host = DEFAULT_HOST
        self._port = DEFAULT_PORT
        self._callsign = DEFAULT_CALLSIGN
        self._server_filter = DEFAULT_SERVER_FILTER
        self._heartbeat = DEFAULT_HEARTBEAT

    @property
    def host(self) -> str:
        return self._host

    @host.setter
    def host(self, host: str) -> None:
        self._host = host

    @property
    def port(self) -> int:
        return self._port

    @port.setter
    def port(self, port: int) -> None:
        self._port = port

    @property
    def callsign(self) -> str:
        return self._callsign

    @callsign.setter
    def callsign(self, callsign: str) -> None:
        self._callsign = callsign

    @property
    def server_filter(self) -> str:
        return self._server_filter

    @server_filter.setter
    def server_filter(self, server_filter: str) -> None:
        self._server_filter = server_filter

    @property
    def heartbeat(self) -> int:
        return self._heartbeat

    @heartbeat.setter
    def heartbeat(self, heartbeat: int) -> None:
        self._heartbeat = heartbeat



import logging
from typing import List

_logger = logging.getLogger(__name__)


def parse_body_message(raw_body: str, strict_mode: bool = False) -> dict:
    if not raw_body:
        raise ValueError("Invalid raw body")

    items: List[str] = raw_body.split(":")
    if len(items) < 3:
        raise ValueError("Invalid message raw body")

    raw_addressee: str = items[1]
    if len(raw_addressee) != 9:
        raise ValueError("Invalid addressee length")

    packet: dict = {
        "addressee": raw_addressee.strip().upper(),
        "message": ":".join(items[2:]).strip()
    }

    message_items: List[str] = packet["message"].split(".")
    if len(message_items) >= 2 and message_items[0].upper() in ["PARM", "UNIT", "EQNS", "BITS"]:
        telemetry_type = message_items[0].strip().upper()
        telemetry_items = (".".join(message_items[1:])).split(",")

        packet.update({
            "telemetry_data": {
                "type": telemetry_type,
                "items": telemetry_items
            }
        })

        if packet["telemetry_data"]["type"] == "EQNS":
            equations: List[dict] = list()

            for pos in range(len(telemetry_items)):
                analog_value_index: int = int(pos / 3)
                if len(equations) < analog_value_index + 1:
                    equations.append(dict())
                equations[analog_value_index]["abc"[pos % 3]] = float(telemetry_items[pos])

            packet["telemetry_data"].update({
                "equations": equations
            })

        elif packet["telemetry_data"]["type"] == "BITS":
            if telemetry_items[0]:
                packet["telemetry_data"].update({
                    "bits": [int(x) == 1 for x in telemetry_items[0]]
                })

            if telemetry_items[1]:
                packet["telemetry_data"].update({
                    "project_name": ",".join(telemetry_items[1:]).strip()
                })

    return packet



import logging
from typing import List

from aioaprs.packets import DATA_TYPE_IDENTIFIER, PacketType
from aioaprs.parsers.message import parse_body_message
from aioaprs.parsers.telemetry import parse_body_telemetry

_logger = logging.getLogger(__name__)


class PacketParser:
    _raw_packet: str
    _packet: dict

    def __init__(self, raw_packet: str) -> None:
        super().__init__()

        self._raw_packet = raw_packet
        self._packet = {}

    def parse(self, strict_mode: bool = False) -> dict:
        if self._raw_packet is None:
            raise ValueError("Raw packet is None")

        if self._raw_packet.startswith("#"):
            return dict()

        self._packet["raw"] = self._raw_packet

        marker_position: int = self._raw_packet.find(":")
        raw_header: str = self._raw_packet[:marker_position]
        raw_body: str = self._raw_packet[marker_position + 1:]

        if not raw_header:
            raise ValueError("Empty header")
        if not raw_body:
            raise ValueError("Empty body")

        self._parse_header(raw_header, strict_mode=strict_mode)
        self._parse_body(raw_body, strict_mode=strict_mode)

        return self._packet

    def _parse_header(self, raw_header: str, strict_mode: bool = False) -> None:
        self._packet["header"] = raw_header

        items: List[str] = raw_header.split(",")
        if len(items) < 2:
            raise ValueError("Invalid header")

        srcdest_items: List[str] = items[0].split(">")
        self._packet["source"] = srcdest_items[0].upper()

        try:
            self._packet["destination"] = srcdest_items[1].upper()
        except IndexError:
            _logger.error(f"No destination parsing header: \"{self._raw_packet}\"")

        self._packet["path"] = items[1:]
        self._packet["via"] = items[-1]

    def _parse_body(self, raw_body: str, strict_mode: bool = False) -> None:
        self._packet["body"] = raw_body

        packet_type_char: str = raw_body[0]

        if packet_type_char in DATA_TYPE_IDENTIFIER:
            packet_type: PacketType = DATA_TYPE_IDENTIFIER[packet_type_char]
            self._packet["type"] = packet_type

        else:
            self._packet["type"] = None

        updated_body: dict = dict()

        if self._packet["type"] == PacketType.TELEMETRY_DATA:
            updated_body = parse_body_telemetry(raw_body, strict_mode=strict_mode)
        elif self._packet["type"] == PacketType.MESSAGE:
            updated_body = parse_body_message(raw_body, strict_mode=strict_mode)

        self._packet.update(updated_body)
